/*!
 * \file sock.c
 * Portable Network Library
 * Socket Module
 */
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#ifdef _WIN32
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#endif
#include <stdlib.h>

#include <pnsock.h>
#ifdef ENABLE_LOG
#include <pnsocklog.h>
#endif

#define	SUCC		(0)
#define FAIL		(-1)

#ifdef _WIN32
#define PNSOCK_ERROR	SOCKET_ERROR
#else
#define PNSOCK_ERROR	(-1)
#endif

#ifdef _WIN32
#define pnsock_sleep(sec)	Sleep((sec) * 1000)
#else
#define pnsock_sleep(sec)	sleep(sec)
#endif

int sock_errno=0;
const int WAIT_SECS = 10;
const int CTS_MAX_TRY = 5;
static int SEND_TIMEOUT = 10;
static int RECV_TIMEOUT = 10;
static char sock_err_str[1024];

int pnsock_startup(void)
{
#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD( 2, 2 );
 
	err = WSAStartup( wVersionRequested, &wsaData );

	if ( err != 0 ) {
	    /* Tell the user that we could not find a usable */
	    /* WinSock DLL.                                  */
	    return -1;
	}
 
	/* Confirm that the WinSock DLL supports 2.2.*/
	/* Note that if the DLL supports versions greater    */
	/* than 2.2 in addition to 2.2, it will still return */
	/* 2.2 in wVersion since that is the version we      */
	/* requested.                                        */
 	if ( LOBYTE( wsaData.wVersion ) != 2 ||
			HIBYTE( wsaData.wVersion ) != 2 ) {
		/* Tell the user that we could not find a usable */
		/* WinSock DLL.                                  */
		WSACleanup( );
		return -1; 
	}

	return 0;
#endif
}

void pnsock_cleanup(void)
{
#ifdef _WIN32
	WSACleanup( );
#endif
}

void pnsock_setopt(int sockfd, int optname, int optval)
{
	switch (optname) {
		case SOCKOPT_SEND_TIMEOUT:
			SEND_TIMEOUT = optval;
			{
				struct timeval tv;
				tv.tv_sec = SEND_TIMEOUT;
				tv.tv_usec = 0;

				setsockopt(sockfd, SOL_SOCKET, SO_SNDTIMEO, (char *)&tv, sizeof(tv));
			}
			break;

		case SOCKOPT_RECV_TIMEOUT:
			RECV_TIMEOUT = optval;
			{
				struct timeval tv;
				tv.tv_sec = RECV_TIMEOUT;
				tv.tv_usec = 0;

				setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv, sizeof(tv));
			}
			break;

		case SOCKOPT_KEEPALIVE:
			setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE, (char *)&optval, sizeof(optval));
			break;
	}
}


unsigned long int conv_hstr2addr(char *dest_addr)
{
	int i;
	struct hostent *host;
	unsigned long int addr=0;
#ifdef INET_ATON
	struct in_addr iaddr;

	if(inet_aton(dest_addr, &iaddr) == 1)  /*  ADDR */
	{
		addr = iaddr.s_addr ;
	} 
	else
#else
	addr = inet_addr(dest_addr);
	if(addr == -1)
#endif
	{  	/*  ּҰ ƴϸ ̸ּ Ȯ */

		host = gethostbyname(dest_addr);  
				/* BSD ֱٹ "10.200.32.218"  ּҵ   */
		if(!host) return -1;   /* none host */

		for(i = 0; i < 4; i++) {  /* 4Ʈ ּҸ (uint)addr ֱ */
			addr = addr << 8;
			addr += (unsigned char)host->h_addr[i];
		} /* end-for */

		addr = htonl(addr) ;
	} /* end-else */

	return addr;
}

/**
 * Portable Network Library / Socket / pnsock_connect
 *
 * connect to given the host address and port.
 * @param dest_addr	Destination address to connect.
 * @param port 		Destination port to connect.
 * @return 			Socket FD if OK, -1 on fail.
 */
int pnsock_connect(char *dest_addr, unsigned short port) 
{
	int i;
	int sock;
	char one=1;
	struct sockaddr_in	serv_addr;
	struct hostent *host;
	unsigned long int addr=0;
#ifdef INET_ATON
	struct in_addr iaddr;

	/* interprets the character string which represents a numeric internet address */
	if(inet_aton(dest_addr, &iaddr) == 1)
	{
		addr = iaddr.s_addr ;
	} 
	else
#else
	addr = inet_addr(dest_addr);
	if(addr == -1)
#endif
	{  	/* 
		 * interprets domain name address if not numeric address.
		 */

		/* Old version doesn't support for interpreting the character string numeric address like "10.200.32.218" */
		host = gethostbyname(dest_addr);  
		if(!host) return -1;   /* none host */

		for(i = 0; i < 4; i++) {  /* 4Ʈ ּҸ (uint)addr ֱ */
			addr = addr << 8;
			addr += (unsigned char)host->h_addr[i];
		}

		addr = htonl(addr) ;
	} /* else */


	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = addr ; /* htonl(addr); */
	serv_addr.sin_port = htons(port);
  
  	for(i = 0; i < CTS_MAX_TRY; i++) {
		if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	      	return -1;
   		}
		setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&one, sizeof(one));
		{
			struct timeval tv;
			tv.tv_sec = SEND_TIMEOUT;
			tv.tv_usec = 0;

			setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, (char*)&tv, sizeof(tv));
			
			tv.tv_sec = RECV_TIMEOUT;
			tv.tv_usec = 0;

			setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, sizeof(tv));
		}
  
    	if(connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == PNSOCK_ERROR)
		{
			pnsock_close((unsigned short)sock);
    	}
		else
		{
			return sock;
		}

		pnsock_sleep(1);
	}

  	return -1;
}

int connui_to(unsigned long int addr, unsigned short port) 
{
	int i;
	int sock;
	char one=1;
	struct sockaddr_in	serv_addr;

	memset(&serv_addr, 0x00, sizeof(serv_addr));

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = addr ; /* htonl(addr); */
	serv_addr.sin_port = htons(port);
  
  	for(i = 0; i < CTS_MAX_TRY; i++) {
		if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	      	return -1;
   		}
		setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&one, sizeof(one));
  
    	if(connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0)
		{
			pnsock_close((unsigned short)sock);
    	}
		else
		{
			return sock;
		}
		pnsock_sleep(1);
	}

  	return -1;
}

/**
 * Portable Network Library / Socket / pnsock_open
 *
 * open the listening port by given the Server address and port.
 * @param serv_addr Server address to open.
 * @param port 		the port to open.
 * @return			Socket File Descriptor if OK, -1 on fail.
 *
 */
int pnsock_open(struct sockaddr_in *serv_addr, unsigned short port)
{
  int flag=0, flag_len=0;
  int sock=0;

   /*
    *  -1 is returned if an error occurs;  otherwise  the  return
	*  value is a descriptor referencing the socket.
	*/
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if(sock == -1) {
	// error log
    return -1;
  }

  // setsockopt:SO_REUSEADDR  added by diver on 2003-12-22
  flag = 1;
  flag_len = sizeof(flag);
  setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, flag_len);

  serv_addr->sin_family = AF_INET;
  serv_addr->sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr->sin_port = htons(port);


  if(bind(sock, (struct sockaddr *)serv_addr, sizeof(*serv_addr)) < 0) {
#ifdef ENABLE_LOG
    pnsocklog_error("bind");
#endif
    pnsock_close((unsigned short)sock);
    return -1;
  }

  if ( listen(sock, 10) == -1 ) return -1;

  return sock;
}

/**
 * Portable Network Library / Socket / pnsock_accept
 *
 * accept client incomming.
 * @param sockfd Listening-socket FD.
 * @param cliaddr Client address.
 * @param clilen Client address buffer's length.
 * @return New Socket FD if OK, -1 on fail.
 */
int pnsock_accept(int sockfd, struct sockaddr_in *cliaddr, int *clilen)
{
	int connfd=0;
	/*
	struct in_addr cli_addr;
	*/

	*clilen = sizeof(struct sockaddr_in);
	connfd = accept (sockfd, (struct sockaddr *)cliaddr, clilen);
	if(connfd == FAIL) {
#ifdef ENABLE_LOG
		pnsocklog_error("accept failed");
#endif
		return FAIL;
	}
	/*
	memset(&cli_addr, 0x00, sizeof(cli_addr));
	cli_addr.s_addr = inet_netof(cliaddr->sin_addr);
	*/
#ifdef ENABLE_LOG
	pnsocklog_info("accepted %s:%d", inet_ntoa(cliaddr->sin_addr), ntohs(cliaddr->sin_port));
#endif

	return connfd;
}

/**
 * Portable Network Library / Socket / pnsock_close
 *
 * close the socket FD.
 * @param sockfd Socket File Descriptor.
 */
void pnsock_close(unsigned short int sockfd)
{
#ifdef ENABLE_LOG
	pnsocklog_info("socket close (fd: %d)", sockfd);
#endif
#ifdef _WIN32
	closesocket(sockfd);
#else
	close(sockfd);
#endif
}

/* =========================================================================== *
 * .Name: sock_err
 * .Description: 
 * .Parameter:
 * .Purpose: 
 * .Assumtions: 
 * .Effects:
 * .Return:
 * ===========================================================================*/
char *
sock_err(void)
{
	return sock_err_str;
}


/*
 * Portable Network Library / Socket / pnsock_send
 *
 * Send data through the socket.
 * @param sockfd	Socket File Descriptor.
 * @param msg		Data.
 * @param size		Data size.
 * @return 			Sent data size if OK, -1 on fail.
 */
int pnsock_send(unsigned int sockfd, void *msg, int size)
{
	int rtn;
	fd_set wfds;
	struct timeval tv;
	int send_size=0;

	sock_errno = 0;
	tv.tv_sec = SEND_TIMEOUT;
	tv.tv_usec = 0;

 	FD_ZERO(&wfds);
	FD_SET(sockfd, &wfds);

	rtn = select(sockfd+1, NULL, &wfds, NULL, &tv);

	/* timeout */
	if ( rtn == 0 ) {
#ifdef ENABLE_LOG
		pnsocklog_warn("%s:%d:select timeout", __FILE__, __LINE__);
#endif
		return 0;
	}
	/* writable */
	if ( rtn > 0 ) {
		if(FD_ISSET(sockfd, &wfds)) { 
#ifdef _WIN32
			send_size = send(sockfd, msg, size, 0);
#else
#ifdef MSG_NOSIGNAL
			send_size = send(sockfd, msg, size, MSG_DONTWAIT | MSG_NOSIGNAL);
#else
			send_size = send(sockfd, msg, size, MSG_DONTWAIT);
#endif
#endif
			if(send_size <= 0) {
				sock_errno = SOCK_DISCONNECTED;
#ifdef ENABLE_LOG
				pnsocklog_error("%s:%d:send error:return value:%d:(errno:%d:%s)", \
						__FILE__, __LINE__, send_size, errno, strerror(errno));
#endif
					return -1;
			}
#ifdef DEBUG
#ifdef ENABLE_LOG
			pnsocklog_debug("SOCKET SEND SIZE: %d", send_size);
			pnsocklog_hex_t("SEND DATA", msg, send_size);
#endif
#endif
			return send_size;
   		}
	}

  	/* error occurred */ 
	if ( rtn == -1 ) {
		/*
		 * select return -1 and set errno as EINTR
		 * when interrupted by signal.
		 */
		if (errno == EINTR) {
#ifdef ENABLE_LOG
			pnsocklog_error("%s:%d:select interrupted", __FILE__, __LINE__);
#endif
		} else {
#ifdef ENABLE_LOG
			pnsocklog_error("%s:%d:select error:(%d)%s", __FILE__, __LINE__, errno, strerror(errno));
#endif
		}
	}

	return send_size;
}

/**
 * Portable Network Library / Socket / pnsock_send_f
 *
 * Send data through the socket until sending complete.
 * Call pnsock_send inside.
 * @param sockfd	Socket File Descriptor.
 * @param msg		Data.
 * @param size		Data size.
 * @return 			Sent data size if OK, -1 on fail.
 */
int pnsock_send_f(int sockfd, void *msg, int size)
{
	int sz=0;
	int send_size=0;
	int sent_size=0;

	send_size = size;

	while(1) {
		sz = pnsock_send(sockfd, (char *)msg + sent_size, send_size);
		if(sz == 0) {
#ifdef ENABLE_LOG
			pnsocklog_print("%s:%d:sock_send:connection reset or timeout", \
							__FILE__, __LINE__);
#endif
			return 0;
		}
		if ( sz < 0 ) {
#ifdef ENABLE_LOG
			pnsocklog_error("%s:%d:sock_send error", \
							__FILE__, __LINE__);
#endif
#ifdef DEBUG
#ifdef ENABLE_LOG
			pnsocklog_debug("sock_send_f sz: %d", sz);
			pnsocklog_debug("size: %d", size);
			pnsocklog_debug("send_size: %d", send_size);
#endif
#endif
			return sz;
		}


		send_size -= sz;
		sent_size += sz;
		if ( send_size == 0 ) break;
	}

	return size;
}

/**
 * Portable Network Library / Socket / pnsock_recv
 *
 * Receive data through the socket.
 * @param sockfd	Socket File Descriptor.
 * @param buf		Buffer to receive.
 * @param sz		Buffer size.
 * @return 			Received bytes size if OK, -1 on fail, 0 on disconnection.
 */
int pnsock_recv(unsigned int sockfd, void *buf, int sz)
{
	int retval;
	fd_set rfds;
	struct timeval tv;
	int rsize=0;

	sock_errno = 0;
	tv.tv_sec = RECV_TIMEOUT;
	tv.tv_usec = 0;
	FD_ZERO(&rfds);
	FD_SET(sockfd, &rfds);

	retval = select(sockfd+1, &rfds, NULL, NULL, &tv);
	if(retval > 0) {
		if(FD_ISSET(sockfd, &rfds)) {
			rsize = recv(sockfd, buf, sz, 0);
			/* timeout or connection reset */
			if(rsize == 0) {
				sock_errno = SOCK_DISCONNECTED;
#ifdef ENABLE_LOG
				pnsocklog_print("%s:%d:connection reset", \
								__FILE__, __LINE__);
#endif
				return 0;
			}
			if (rsize < 0) {
#ifdef ENABLE_LOG
				pnsocklog_error("%s:%d:recv error:return value:%d(errno:%d:%s)", \
					__FILE__, __LINE__, rsize, errno, strerror(errno));
#endif
				return -1;
			}
#ifdef DEBUG
#ifdef ENABLE_LOG
			pnsocklog_debug("SOCK RECV:size:%d", rsize);
#endif
#endif
		}
#ifdef DEBUG
#ifdef ENABLE_LOG
		pnsocklog_hex_t("RECV DATA", buf, rsize);
#endif
#endif
	}
	if(retval == 0) {
#ifdef ENABLE_LOG
		pnsocklog_print("%s:%d:select timeout", __FILE__, __LINE__);
#endif
		return 0;
	}
	if(retval == -1) {
#ifdef ENABLE_LOG
		pnsocklog_error("%s:%d:select error:(%d)%s", \
				__FILE__, __LINE__, errno, strerror(errno));
#endif
		return FAIL;
	}

	return rsize;
} 


/**
 * Portable Network Library / Socket / pnsock_recv_f
 *
 * Try getting data through the socket until receiving complete.
 * @param sockfd	Socket File Descriptor.
 * @param buf		Buffer to receive.
 * @param len		Buffer size.
 * @return 			Received bytes size if OK, -1 on fail, 0 on disconnection.
 */
int pnsock_recv_f(unsigned int sockfd, void *buf, int len)
{
	int sz;
	int recv_sz;

	recv_sz = len;

	while(1) {
		sz = pnsock_recv(sockfd, buf, recv_sz);
		if (sz == 0) {
#ifdef ENABLE_LOG
			pnsocklog_print("%s:%d:pnsock_recv:timeout or connection reset", \
							__FILE__, __LINE__);
#endif
			return 0;
		}
		if ( sz < 0 ) {
#ifdef ENABLE_LOG
			pnsocklog_error("%s:%d:pnsock_recv error", __FILE__, __LINE__);
#endif
#ifdef DEBUG
#ifdef ENABLE_LOG
			pnsocklog_debug("pnsock_recv_f sz: %d", sz);
			pnsocklog_debug("len: %d", len);
			pnsocklog_debug("recv_sz: %d", recv_sz);
#endif
#endif
			return sz;
		}

		recv_sz -= sz;
		if ( recv_sz == 0 ) break;
	}

	return len;
}

/**
 * Portable Network Library / Socket / pnsock_recv_f
 *
 * Similar with pnsock_recv_f except allocating memory and returning 
 * the allocated memory.
 * @param rcv_data	Pointer to allocate memory.
 * @param sockfd	Socket File Descriptor.
 * @param length	Packet size.
 * @return 			Received bytes size if OK, -1 on fail, 0 on disconnection.
 */
int pnsock_recv_alloc_f(void **rcv_data, int sockfd, int length)
{
	int sz;
	int data_size=0;
	int recv_length=0;
	int arg_recv_len;
	char rcvbuf[2048];
	void *ptr=NULL;

	recv_length = length;

	while(1) {
		arg_recv_len = recv_length;
		if(recv_length > sizeof(rcvbuf)) arg_recv_len = sizeof(rcvbuf);
		memset(rcvbuf, 0x00, sizeof(rcvbuf));

		/* read socket */
		sz = pnsock_recv(sockfd, rcvbuf, arg_recv_len);
		if (sz == 0) {
#ifdef ENABLE_LOG
			pnsocklog_warn("sock_recv:timeout or connection reset");
#endif
			return 0;
		}
		if ( sz < 0 ) {
#ifdef ENABLE_LOG
			pnsocklog_error("%s:%d:sock_recv error", __FILE__, __LINE__);
#endif
#ifdef DEBUG
#ifdef ENABLE_LOG
			pnsocklog_debug("data_size: %d", data_size);
			pnsocklog_debug("recv_length: %d", recv_length);
			pnsocklog_debug("arg_recv_len: %d", arg_recv_len);
#endif
#endif
			if(ptr) { free(ptr); ptr = NULL; }
			return sz;
		}

		ptr = realloc(ptr, (data_size + sz));
		if(!ptr) {
#ifdef ENABLE_LOG
			pnsocklog_error("%s:%d:realloc error", __FILE__, __LINE__);
#endif
			return FAIL;
		}
		memcpy( (((char *)ptr)+data_size) , rcvbuf , sz );

		data_size += sz;
		recv_length -= sz;
		if ( recv_length == 0 ) break;
	}

	*rcv_data = ptr;

	return data_size;
}


int sock_getc(unsigned int sock)
{
  struct timeval tv;
  fd_set rfds;
  int sz;
  unsigned char ch;

  tv.tv_sec = WAIT_SECS;
  tv.tv_usec = 0;

  FD_ZERO(&rfds);
  FD_SET(sock, &rfds);

  if(select(sock+1, &rfds, NULL, NULL, &tv) > 0) {
    sz = recv(sock, &ch, 1, 0);
    if(sz <= 0) return -1;

    return ch;
  } 

  return -1;
}

int sock_recvline(int sock, char *buf, int leng)
/*
 *	sock	: socket's file descriptor
 *	buf		: the buffer be saved
 *	leng	: size of buffer
 *	retval	: the number of characters recevied a line
 */
{
  int ch, i;

  for(i = 0; i < leng; i++) {
    ch = sock_getc(sock);

    if(ch == '\n') {
      buf[i] = 0;
      return i;
    }
    else if(ch <= 0) 
      return -1;
    else buf[i] = (char)ch;
  }

  return i;
}

void sock_error_print(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);

#ifdef ENABLE_LOG
	pnsocklog_error(fmt, args);
#endif

	va_end( args );
}
